/*
 * OpenCL Info
 * Copyright (C) 2009 cmf
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "main.h"

void init_platform() {
	try {
		platform = new cl::Platform();
		platform->get(&platforms);
		
		cout << platforms.size() << " OpenCL platform" << (platforms.size() > 1 ? "s" : "") << " found!" << endl << endl;
		
		// get platform info and get devices from the first platform
		unsigned int platform_counter = 0;
		string platform_profile, platform_version, platform_name, platform_vendor, platform_extensions;
		for(vector<cl::Platform>::iterator piter = platforms.begin(); piter != platforms.end(); piter++) {
			cout << "[Platform " << platform_counter << "]" << endl;
			piter->getInfo(CL_PLATFORM_PROFILE, &platform_profile);
			piter->getInfo(CL_PLATFORM_VERSION, &platform_version);
			piter->getInfo(CL_PLATFORM_NAME, &platform_name);
			piter->getInfo(CL_PLATFORM_VENDOR, &platform_vendor);
			piter->getInfo(CL_PLATFORM_EXTENSIONS, &platform_extensions);
			
			cout << TAB << "Name: " << TAB << TAB << platform_name << endl;
			cout << TAB << "Vendor: " << TAB << platform_vendor << endl;
			cout << TAB << "Version: " << TAB << platform_version << endl;
			cout << TAB << "Profile: " << TAB << platform_profile << endl;
			if(platform_extensions.length() > 0) cout << TAB << "Extensions: " << TAB << TAB << TAB << platform_extensions << endl;
			
			// since there only will be one platform ... probably :>
			if(platform_counter == 0) piter->getDevices(CL_DEVICE_TYPE_ALL, &devices);
			
			cout << endl;
			platform_counter++;
		}
		
		cout << endl;
	}
	catch(cl::Error cl_err) {
		cout << "Error: " << "init(): " << cl_err.what() << " " << cl_err.err() << "!" << endl;
	}
	catch(...) {
		cout << "Error: " << "init(): unknown error!" << endl;
	}
}

void init(bool ogl_sharing) {
	if(platforms.size() == 0) return;
	
	cl_int err;
	try {
		if(ogl_sharing) {
			cl_context_properties cl_properties[] = {
				CL_CONTEXT_PLATFORM, (cl_context_properties)platforms[0](),
				CL_CONTEXT_PROPERTY_USE_CGL_SHAREGROUP_APPLE, (cl_context_properties)CGLGetShareGroup(CGLGetCurrentContext()),
				0
			};
			
			// add all cpu devices to the initial shared devices list (don't know why, but if gpu devices are
			// added, the context creation will just fail ... so no gpu+gpu+cpu context is currently possible :/)
			vector<cl::Device> shared_devices;
			for(vector<cl::Device>::iterator dev_iter = devices.begin(); dev_iter != devices.end(); dev_iter++) {
				if(dev_iter->getInfo<CL_DEVICE_TYPE>() & CL_DEVICE_TYPE_CPU &&
				   dev_iter->getInfo<CL_DEVICE_EXTENSIONS>().find("cl_APPLE_gl_sharing") != string::npos) {
					shared_devices.push_back(*dev_iter);
				}
			}
			
			context = new cl::Context(shared_devices, cl_properties, NULL, NULL, &err);
			
			devices.clear();
			devices = context->getInfo<CL_CONTEXT_DEVICES>();
		}
		else {
			// create a opencl context for all available/capable devices
			context = new cl::Context(CL_DEVICE_TYPE_ALL, NULL, NULL, NULL, &err);
		}
		
		cout << (ogl_sharing ? "[shared OpenCL+OpenGL Context]" : "[OpenCL-only Context]") << endl;
		cout << devices.size() << " OpenCL device" << (devices.size() > 1 ? "s" : "") << " found!" << endl << endl;
		
		unsigned dev_counter = 0;
		string dev_name;
		string dev_vendor;
		string dev_version;
		string dev_driver_version;
		string dev_extensions;
		cl_device_type dev_type;
		string dev_type_str;
		unsigned int dev_units;
		unsigned int dev_clock;
		size_t dev_work_group_size;
		unsigned long int dev_global_mem_size;
		unsigned long int dev_local_mem_size;
		unsigned long int dev_global_mem_cache_size;
		unsigned int dev_global_mem_cache_line_size;
		bool dev_available;
		for(vector<cl::Device>::iterator dev_iter = devices.begin(); dev_iter != devices.end(); dev_iter++) {
			// gather device info
			dev_name = dev_iter->getInfo<CL_DEVICE_NAME>();
			dev_vendor = dev_iter->getInfo<CL_DEVICE_VENDOR>();
			dev_version = dev_iter->getInfo<CL_DEVICE_VERSION>();
			dev_driver_version = dev_iter->getInfo<CL_DRIVER_VERSION>();
			dev_extensions = dev_iter->getInfo<CL_DEVICE_EXTENSIONS>();
			dev_type = dev_iter->getInfo<CL_DEVICE_TYPE>();
			dev_type_str = "";
			if(dev_type & CL_DEVICE_TYPE_CPU) dev_type_str += "CPU ";
			if(dev_type & CL_DEVICE_TYPE_GPU) dev_type_str += "GPU ";
			if(dev_type & CL_DEVICE_TYPE_ACCELERATOR) dev_type_str += "Accelerator ";
			if(dev_type & CL_DEVICE_TYPE_DEFAULT) dev_type_str += "Default ";
			
			dev_available = dev_iter->getInfo<CL_DEVICE_AVAILABLE>();
			dev_units = dev_iter->getInfo<CL_DEVICE_MAX_COMPUTE_UNITS>();
			dev_clock = dev_iter->getInfo<CL_DEVICE_MAX_CLOCK_FREQUENCY>();
			dev_global_mem_size = dev_iter->getInfo<CL_DEVICE_GLOBAL_MEM_SIZE>();
			dev_local_mem_size = dev_iter->getInfo<CL_DEVICE_LOCAL_MEM_SIZE>();
			dev_work_group_size = dev_iter->getInfo<CL_DEVICE_MAX_WORK_GROUP_SIZE>();
			
			dev_global_mem_cache_size = dev_iter->getInfo<CL_DEVICE_GLOBAL_MEM_CACHE_SIZE>();
			dev_global_mem_cache_line_size = dev_iter->getInfo<CL_DEVICE_GLOBAL_MEM_CACHELINE_SIZE>();
			
			// print info
			cout << "[Device " << dev_counter << "]" << endl;
			cout << TAB << "Name: " << TAB << TAB << TAB << dev_name << endl;
			cout << TAB << "Vendor: " << TAB << TAB << dev_vendor << endl;
			cout << TAB << "Type: " << TAB << TAB << TAB << dev_type_str << endl;
			cout << TAB << "Device Version: " << TAB << dev_version << endl;
			cout << TAB << "Driver Version: " << TAB << dev_driver_version << endl;
			cout << TAB << "Compute Units: " << TAB << TAB << dev_units << endl;
			cout << TAB << "Work Group Size: " << TAB << dev_work_group_size << endl;
			cout << TAB << "Clock: " << TAB << TAB << TAB << dev_clock << " MHz" << endl;
			
			if(dev_type & CL_DEVICE_TYPE_CPU) {
				cout << TAB << "Global Memory (Total): " << TAB << cpu_memory << " MB" << endl;
				cout << TAB << "Global Memory (Host): " << TAB << (dev_global_mem_size / 1024ul / 1024ul) << " MB" << endl;
				cout << TAB << "Global Memory (PCIe): " << TAB << (cpu_memory - (dev_global_mem_size / 1024ul / 1024ul)) << " MB" << endl;
			}
			else cout << TAB << "Global Memory: " << TAB << TAB << (dev_global_mem_size / 1024ul / 1024ul) << " MB" << endl;
			
			cout << TAB << "Local Memory: " << TAB << TAB << ((float)dev_local_mem_size / 1024.0f) << " KB" << endl;
			cout << TAB << "Cache Size: " << TAB << TAB << ((float)dev_global_mem_cache_size / 1024.0f) << " KB" << endl;
			cout << TAB << "Cache Line Size: " << TAB << (dev_global_mem_cache_line_size) << " Bytes" << endl;
			cout << TAB << "Available: " << TAB << TAB << (dev_available ? "Yes" : "No") << endl;
			cout << TAB << "Double-Precision: " << TAB << ((dev_extensions.find("cl_khr_fp64", 0) != string::npos) ? "Yes" : "No") << endl;
			cout << TAB << "Extensions: " << endl;
			
			size_t ext_pos = 0, last_pos = 0;
			while((ext_pos = dev_extensions.find(" ", last_pos)) != string::npos || last_pos < dev_extensions.length()) {
				if(ext_pos == string::npos) ext_pos = dev_extensions.length();
				cout << TAB << TAB << TAB << TAB << dev_extensions.substr(last_pos, ext_pos - last_pos) << endl;
				last_pos = ext_pos+1;
			}
			
			cout << endl;
			dev_counter++;
		}
	}
	catch(cl::Error cl_err) {
		cout << "Error: " << "init(): " << cl_err.what() << " " << cl_err.err() << "!" << endl;
	}
	catch(...) {
		cout << "Error: " << "init(): unknown error!" << endl;
	}
}

int main(int argc, char* argv[]) {
	// get cpu memory size
	string system_out = "";
	char buffer[512];
	memset(buffer, 0, 512);
	FILE* sys_pipe = popen("sysctl hw.memsize", "r");
	while(fgets(buffer, 512, sys_pipe) != NULL) {
		system_out += buffer;
		memset(buffer, 0, 512);
	}
	pclose(sys_pipe);
	
	size_t split = system_out.find(" ");
	cpu_memory = strtoul(system_out.substr(split+1, system_out.length()-split).c_str(), NULL, 10) / 1024 / 1024;
	
	if(argc > 1 && strcmp(argv[1], "-version") == 0) {
		cout << "OpenCL Info " << OCLINFO_MAJOR_VERSION << "." << OCLINFO_MINOR_VERSION << " (" << OCLINFO_BUILD_DATE << " " << OCLINFO_BUILD_TIME << ")" << endl << endl;
	}
	
	// opencl info
	try {
		init_platform();
		
		init(false);
		
		create_ogl_context();
		init(true);
		destroy_ogl_context();
	}
	catch (cl::Error err) {
		std::cerr << "Error: " << err.what() << "(" << err.err() << ")" << std::endl;
	}
	
	return 0;
}
